"""Just contains a singleton class. Pretty useful. Class taken from taurex 3. """

class Singleton(object):
    """
    A singleton for your usage. When inheriting do not implement __init__ instead
    override :func:`init`
    
    
    """
    def __new__(cls, *args, **kwds):
        it = cls.__dict__.get("__it__")
        if it is not None:
            return it
        cls.__it__ = it = object.__new__(cls)
        it.init(*args, **kwds)
        return it

    def init(self, *args, **kwds):
        """ Override to act as an init """
        pass
    


# -*- coding: utf-8 -*-
"""
@author: jeremy leconte
"""
import os
import array
import pickle
import h5py
from astropy.io import fits
import numpy as np
import astropy.units as u
from .data_table import Data_table
from .util.interp import rm_molec, gauss_legendre
from .util.filenames import _read_array, _read_exorem_k_array
from .util.cst import nemesis_hitran_id_numbers

class Ktable_io(Data_table):
    """A class to handle the input-output methods of the :class:`~exo_k.ktable.Ktable` class.
    """

    def read_hdf5(self, filename=None, mol=None):
        """Initializes k coeff table and supporting data from an hdf5 file
        (compatible with Exomol format)

        Parameters
        ----------
            file : str
                Name of the input hdf5 file
        """
        if (filename is None or not filename.lower().endswith(('.hdf5', '.h5'))):
            raise RuntimeError("You should provide an input hdf5 file")
        with h5py.File(filename, 'r') as f:
            if 'mol_name' in f:
                self.mol=f['mol_name'][()]
            elif 'mol_name' in f.attrs:
                self.mol=f.attrs['mol_name']
            else:
                if mol is not None:
                    self.mol=mol
                else:
                    self.mol=os.path.basename(filename).split(self._settings._delimiter)[0]
            if isinstance(self.mol, np.ndarray): self.mol=self.mol[0]
            if isinstance(self.mol, bytes): self.mol=self.mol.decode('UTF-8')
            if 'method' in f:
                self.sampling_method=f['method'][()][0]
                if isinstance(self.sampling_method, bytes):
                    self.sampling_method=self.sampling_method.decode('UTF-8')
            if 'DOI' in f:
                self.DOI=f['DOI'][()][0]
                if isinstance(self.DOI, bytes): self.DOI=self.DOI.decode('UTF-8')
            self.wns=f['bin_centers'][...]
            self.wnedges=f['bin_edges'][...]
            if 'units' in f['bin_edges'].attrs:
                self.wn_unit=f['bin_edges'].attrs['units']
            else:
                if 'units' in f['bin_centers'].attrs:
                    self.wn_unit=f['bin_centers'].attrs['units']
            self.kdata=f['kcoeff'][...]
            self.kdata_unit=f['kcoeff'].attrs['units']
            self.tgrid=f['t'][...]
            self.pgrid=f['p'][...]
            self.logpgrid=np.log10(self.pgrid)
            self.p_unit=f['p'].attrs['units']
            if 'weights' in f.keys():
                self.weights=f['weights'][...]
            else:
                raise RuntimeError('No weights keyword. This file is probably a cross section file.')
            self.ggrid=f['samples'][...]
            self.gedges=np.insert(np.cumsum(self.weights),0,0.)
            self.logk=False
        self.Np,self.Nt,self.Nw,self.Ng=self.kdata.shape

    def write_hdf5(self, filename, compression="gzip", compression_level=9,
        kdata_unit=None, p_unit=None, exomol_units=False):
        """Saves data in a hdf5 format

        Parameters
        ----------
            filename: str
                Name of the file to be created and saved
            exomol_units: bool (optional)
                If True, data are converted back to
                cm^2 and bar units before being written.
        """
        dt = h5py.string_dtype(encoding='utf-8')
        fullfilename=filename
        if not filename.lower().endswith(('.hdf5', '.h5')):
            fullfilename=filename+'.h5'
        with h5py.File(fullfilename, 'w') as f:
            f.create_dataset("temperature_grid_type", (1,), data='regular', dtype=dt)
            f.create_dataset("key_iso_ll", (1,), data=self.isotopolog_id)
            f.create_dataset("mol_mass", (1,), data=self.molar_mass*1000.)
            f["mol_mass"].attrs["units"] = 'AMU'
            if exomol_units:
                kdata_unit='cm^2/molecule'
                p_unit='bar'
            if kdata_unit is not None:
                conv_factor=u.Unit(rm_molec(self.kdata_unit)).to(u.Unit(rm_molec(kdata_unit)))
                data_to_write=self.kdata*conv_factor
                f.create_dataset("kcoeff", data=data_to_write,
                    compression=compression, compression_opts=compression_level)
                f["kcoeff"].attrs["units"] = kdata_unit
            else:
                f.create_dataset("kcoeff", data=self.kdata,
                    compression=compression, compression_opts=compression_level)
                f["kcoeff"].attrs["units"] = self.kdata_unit
            f.create_dataset("method", (1,), data=self.sampling_method, dtype=dt)
            f.create_dataset("samples", data=self.ggrid,
                compression=compression, compression_opts=compression_level)
            f.create_dataset("weights", data=self.weights,
                compression=compression, compression_opts=compression_level)
            f.create_dataset("ngauss", data=self.Ng)
            f.create_dataset("bin_centers", data=self.wns,
                compression=compression, compression_opts=compression_level)
            f.create_dataset("bin_edges", data=self.wnedges,
                compression=compression, compression_opts=compression_level)

            # where most of the data is actually written
            self.write_hdf5_common(f, compression=compression, compression_level=compression_level,
            p_unit=p_unit)

    def read_LMDZ(self, path=None, res=None, band=None, mol=None):
        """Initializes k coeff table and supporting data from a .dat file in a gcm friendly format.
        
        Units are assumed to be cm^2 for kdata and mbar for pressure. 

        Parameters
        ----------
            path: str
                Name of the directory with the various input files
            res: str
                "IRxVI" where IR and VI are the numbers of bands
                in the infrared and visible of the k table to load.
            band: str
                "IR" or "VI" to specify which band to load.
            mol: str
                Name of the molecule to be saved in the Ktable object. 
        """        
        if (path is None) or (res is None): \
            raise TypeError("You should provide an input directory name and a resolution")

        self.filename=path
        if mol is not None:
            self.mol=mol
        else:
            self.mol=os.path.basename(self.filename).split(self._settings._delimiter)[0]

        self.weights=np.loadtxt(os.path.join(path,'g.dat'),skiprows=1)[:-1]
            # we remove the last point that is always zero.
            # in the gcm this last point is intended to take care of continuum
        self.Ng=self.weights.size
        self.gedges=np.insert(np.cumsum(self.weights),0,0.)
        self.ggrid=(self.gedges[1:]+self.gedges[:-1])*0.5

        self.p_unit='mbar'
        self.logpgrid=np.loadtxt(os.path.join(path,'p.dat'),skiprows=1)*1.
        self.Np=self.logpgrid.size
        self.pgrid=10**self.logpgrid

        self.tgrid=np.loadtxt(os.path.join(path,'T.dat'),skiprows=1)
        self.Nt=self.tgrid.size

        if band is None:
            raw=np.loadtxt(os.path.join(path,res,'narrowbands.in'), skiprows=1, unpack=True)
        else:
            raw=np.loadtxt(os.path.join(path,res,'narrowbands_'+band+'.in'), \
                skiprows=1, unpack=True)
        self.wnedges=np.append(raw[0],raw[1,-1])
        self.wns=(self.wnedges[1:]+self.wnedges[:-1])*0.5
        self.Nw=self.wns.size
        
        self.kdata_unit='cm^2/molecule'
        if band is None:
            file_to_load=os.path.join(path,res,'corrk_gcm.dat')
        else:
            file_to_load=os.path.join(path,res,'corrk_gcm_'+band+'.dat')        
        tmp=np.loadtxt(file_to_load).reshape((self.Nt,self.Np,self.Nw,self.Ng+1),order='F')
        self.kdata=tmp[:,:,:,:-1].transpose((1,0,2,3)) 
        # also removing the last g point which is equal to 0.
        self.logk=False        
        return None

    def write_LMDZ(self, path, band='IR', fmt='%22.15e', write_only_metadata=False):
        """Saves data in a LMDZ friendly format.
        
        The gcm requires p in mbar and kdata in cm^2/molec.
        The conversion is done automatically.

        Parameters
        ----------
            path: str
                Name of the directory to be created and saved,
                the one that will contain all the necessary files
            band: str
                The band you are computing: 'IR' or 'VI'
            fmt: str
                Fortran format for the corrk file. 
        """
        try:
            os.mkdir(path)
        except FileExistsError:
            print('Directory was already there '+path)
        file = open(os.path.join(path,'p.dat'), "w")
        file.write(str(self.Np)+'\n')
        lp_to_write=self.logpgrid+np.log10(u.Unit(self.p_unit).to(u.Unit('mbar')))
        for lp in lp_to_write:
            file.write(str(lp)+'\n')
        file.close()

        file = open(os.path.join(path,'T.dat'), "w")
        file.write(str(self.Nt)+'\n')
        for t in self.tgrid:
            file.write(str(t)+'\n')
        file.close()

        file = open(os.path.join(path,'g.dat'), "w")
        file.write(str(self.Ng+1)+'\n')
        for g in self.weights:
            file.write(str(g)+'\n')
        file.write(str(0.)+'\n')
        file.close()

        dirname=os.path.join(path,band+str(self.Nw))
        try:
            os.mkdir(dirname)
        except FileExistsError:
            print('Directory was already there: '+dirname)

        file = open(os.path.join(dirname,'narrowbands_'+band+'.in'), "w")
        file.write(str(self.Nw)+'\n')
        for iw in range(self.Nw):
            file.write(str(self.wnedges[iw])+' '+str(self.wnedges[iw+1])+'\n')
        file.close()

        if not write_only_metadata:
            #file = open(dirname+'/corrk_gcm_IR.in', "w")
            data_to_write=self.kdata.transpose((1,0,2,3)).flatten(order='F')
            data_to_write=data_to_write*u.Unit(rm_molec(self.kdata_unit)).to(u.Unit('cm^2'))
            data_to_write=np.append(data_to_write, \
                np.zeros(self.Np*self.Nt*self.Nw)).reshape((1,self.Np*self.Nt*self.Nw*(self.Ng+1)))
            np.savetxt(os.path.join(dirname,'corrk_gcm_'+band+'.dat'),data_to_write,fmt=fmt)

    def read_nemesis(self, filename=None, mol=None):
        """Initializes k coeff table and supporting data from a Nemesis binary file (.kta)

        Parameters
        ----------
            file: str
                Name of the input Nemesis binary file.
            mol: str, optional
                Name of the molecule.
        """
        if (filename is None or not filename.lower().endswith('.kta')):
            raise RuntimeError("You should provide an input nemesis (.kta) file")
        
        self.Np, self.Nt, self.Nw, self.Ng, \
            self.pgrid, self.tgrid, self.wns, \
            self.ggrid, self.weights, self.kdata, \
            self.mol, self.isotopolog_id = read_nemesis_binary(filename)

        if self.mol is None:
            if mol is not None:
                self.mol=mol
            else:
                self.mol=os.path.basename(self.filename).split(self._settings._delimiter)[0]

        self.logpgrid=np.log10(self.pgrid)
        self.wnedges=np.concatenate(  \
            ([self.wns[0]],(self.wns[:-1]+self.wns[1:])*0.5,[self.wns[-1]]))
        self.gedges=np.insert(np.cumsum(self.weights),0,0.)
        self.p_unit='bar'
        self.kdata_unit='cm^2/molecule'

    def write_nemesis(self, filename):
        """Saves data in a nemesis format.

        Based on a routine provided by K. Chubb.

        Parameters
        ----------
            filename: str
                Name of the file to be created and saved
        """
        int_format = np.int32
        float_format = np.float32
        fullfilename=filename
        if not filename.lower().endswith(('.kta')):
            fullfilename=filename+'.kta'
        with open(fullfilename,'wb') as o:
            o.write(int_format(11+2*self.Ng+2+self.Np+self.Nt+self.Nw).tostring())
            #int for irec0 (5221)
            o.write(int_format(self.Nw).tostring()) #int
            o.write(float_format(self.wls[-1]).tostring()) #float for VMIN (0.30015)
            o.write(float_format(-1.).tostring()) #float
            o.write(float_format(0.).tostring()) #float for FWHM
            o.write(int_format(self.Np).tostring())
            o.write(int_format(self.Nt).tostring())
            o.write(int_format(self.Ng).tostring())
            try:
                mol_id=nemesis_hitran_id_numbers[self.mol]
            except:
                mol_id=0
            o.write(int_format(mol_id).tostring()) #IDGAS1 FROM HITRAN
            o.write(int_format(self.isotopolog_id).tostring())
            o.write(float_format(self.ggrid).tostring())
            o.write(float_format(self.weights).tostring())
            o.write(float_format(0.).tostring()) #float for 0
            o.write(float_format(0.).tostring()) #float for 0
            conv_factor=u.Unit(self.p_unit).to(u.Unit('bar'))
            o.write(float_format(self.pgrid).tostring())
            o.write(float_format(self.tgrid).tostring())
            o.write(float_format(self.wls[::-1]).tostring())
            data_to_write=self.kdata[:,:,::-1,:].transpose(2,0,1,3)
            conv_factor=u.Unit(rm_molec(self.kdata_unit)).to(u.Unit('cm^2'))*1.e20
            data_to_write=data_to_write*conv_factor
            o.write(float_format(data_to_write).tostring())
    
    def read_exorem(self, filename, mol=None):
        """Reads data in an ExoREM .dat format

        Parameters
        ----------
            filename: str
                Name of the input file.
            mol: str, optional
                Name of the molecule.
        """
        if self.mol is None:
            if mol is not None:
                self.mol=mol
            else:
                self.mol=os.path.basename(self.filename).split(self._settings._delimiter)[0]
        with open(filename, 'r') as file:
            tmp = file.readline().split()
            self.Np=int(tmp[0])
            self.Nt=int(tmp[1])
            self.Nw=int(tmp[2])
            self.Ng=int(tmp[-1])
            #print(self.Np,self.Nt,self.Nw,self.Ng)
            self.pgrid=_read_array(file, self.Np,  N_per_line=5, revert=True)
            self.logpgrid=np.log10(self.pgrid)
            self.p_unit='Pa'
            for _ in range(self.Np):
                self.tgrid=_read_array(file, self.Nt,  N_per_line=5)
            self.ggrid=_read_array(file, self.Ng,  N_per_line=5, revert=False)
            self.weights=_read_array(file, self.Ng,  N_per_line=5, revert=False)
            self.gedges=np.concatenate(([0],np.cumsum(self.weights)))

            self.kdata=np.zeros((self.Np, self.Nt, self.Nw, self.Ng))
            self.wns=np.zeros((self.Nw))
            #read_new_line=True
            tmp = file.readline().split()
            for iW in range(self.Nw):
                self.wns[iW]=float(tmp[0])
                #if iW=self.Nw-1: read_new_line=False
                for iP in range(self.Np):
                    for iT in range(self.Nt):
                        self.kdata[-iP-1,iT,iW]=_read_exorem_k_array(file, self.Ng)
                        #if read_new_line: tmp = file.readline().split()
                        tmp = file.readline().split()
            self.kdata_unit='cm^2/molec'
            self.wnedges=np.concatenate( \
                ([self.wns[0]],(self.wns[:-1]+self.wns[1:])*0.5,[self.wns[-1]]))
                
    def read_arcis(self, filename=None, mol=None):
        """Initializes k coeff table and supporting data from an ARCI fits file (.fits)

        Parameters
        ----------
            file: str
                Name of the input fits file.
            mol: str, optional
                Name of the molecule.
        """
        with fits.open(filename) as f:
            self.kdata=f[0].data[:,:,:,::-1] #reverse spectral axis
            self.kdata=np.transpose(self.kdata, axes=(0,1,3,2))
            self.kdata_unit='cm^2/molecule'
            self.Np, self.Nt, self.Nw, self.Ng = self.kdata.shape
            self.tgrid=f[1].data
            self.pgrid=f[2].data
            self.logpgrid=np.log10(self.pgrid)
            self.p_unit='bar'
            self.wns=1./f[3].data[::-1] #data given in wavelength in cm
            self.wn_unit='cm^-1'
            wns_max=1./f[0].header['L_MIN']
            wns_min=1./f[0].header['L_MAX']
            self.wnedges=np.concatenate( \
                ([wns_min],(self.wns[:-1]+self.wns[1:])*0.5,[wns_max]))
            self.weights,self.ggrid,self.gedges=gauss_legendre(self.Ng)

        if mol is not None:
            self.mol=mol
        else:
            self.mol=os.path.basename(filename).split(self._settings._delimiter)[0]

    def write_arcis(self, filename):
        """Saves data in an ARCIS fits format.

        Parameters
        ----------
            filename: str
                Name of the file to be created and saved
        """
        hdr = fits.Header()
        p_conv_fac=u.Unit(self.p_unit).to(u.Unit('bar'))
        hdr['TMIN'] = self.tgrid[0]
        hdr['TMAX'] = self.tgrid[-1]
        hdr['PMIN'] = self.pgrid[0]*p_conv_fac
        hdr['PMAX'] = self.pgrid[-1]*p_conv_fac
        hdr['L_MIN'] = 1./self.wns[-1]
        hdr['L_MAX'] = 1./self.wns[0]
        hdr['NT'] = self.Nt
        hdr['NP'] = self.Np
        hdr['NLAM'] = self.Nw
        hdr['NG'] = self.Ng
        data_to_write=self.kdata[:,:,::-1,:].transpose(0,1,3,2)
        conv_factor=u.Unit(rm_molec(self.kdata_unit)).to(u.Unit('cm^2'))
        hdu0=fits.PrimaryHDU(data_to_write*conv_factor, header=hdr)
        hdu1=fits.ImageHDU(self.tgrid)
        hdu2=fits.ImageHDU(self.pgrid*p_conv_fac)
        hdu3=fits.ImageHDU(1./self.wns[::-1])
        hdul = fits.HDUList([hdu0, hdu1, hdu2, hdu3])

        fullfilename=filename
        if not filename.lower().endswith(('.fits')):
            fullfilename=filename+'.fits'
        hdul.writeto(fullfilename, overwrite=True)

    def read_pickle(self, filename=None):
        """Initializes k coeff table and supporting data from an Exomol pickle file

        Parameters
        ----------
            filename: str
                Name of the input pickle file
            mol: str, optional
                Force the name of the molecule
        """
        if filename is None: raise RuntimeError("You should provide an input pickle filename")
        pickle_file=open(filename,'rb')
        raw=pickle.load(pickle_file, encoding='latin1')
        pickle_file.close()
        
        self.mol=raw['name']
        if self.mol=='H2OP': self.mol='H2O'

        self.pgrid=raw['p']
        self.logpgrid=np.log10(self.pgrid)
        self.tgrid=raw['t']
        self.wns=raw['bin_centers']
        self.wnedges=raw['bin_edges']
        if 'weights' in raw.keys():
            self.weights=raw['weights']
        else:
            raise RuntimeError('No weights keyword. This file is probably a cross section file.')
        self.ggrid=raw['samples']
        self.gedges=np.insert(np.cumsum(self.weights),0,0.)
        self.kdata=raw['kcoeff']
    
        if 'p_unit' in raw.keys():
            self.p_unit=raw['p_unit']
        else:
            self.p_unit='bar'

        if 'kdata_unit' in raw.keys():
            self.kdata_unit=raw['kdata_unit']
        else:
            self.kdata_unit='cm^2/molec'

        if 'wn_unit' in raw.keys(): self.wn_unit=raw['wn_unit']

        self.Np,self.Nt,self.Nw,self.Ng=self.kdata.shape

    def write_pickle(self, filename):
        """Saves data in a pickle format

        Parameters
        ----------
            filename: str
                Name of the file to be created and saved
        """
        fullfilename=filename
        if not filename.lower().endswith('.pickle'): fullfilename=filename+'.pickle'
        pickle_file=open(fullfilename,'wb')
        dictout={'name':self.mol,
                 'p':self.pgrid,
                 'p_unit':self.p_unit,
                 't':self.tgrid,
                 'bin_centers':self.wns,
                 'bin_edges':self.wnedges,
                 'wn_unit':self.wn_unit,
                 'weights':self.weights,
                 'samples':self.ggrid,
                 'kcoeff':self.kdata,
                 'kdata_unit':self.kdata_unit}
        #print(dictout)
        pickle.dump(dictout,pickle_file,protocol=-1)
        pickle_file.close()

    def spectrum_to_plot(self, p=1.e-5, t=200., x=1., g=None):
        """Dummy function to be defined in inheriting classes
        """
        raise NotImplementedError()

def read_nemesis_binary(filename):
    """reads a nemesis binary file.
    """
    f = open(filename, mode='rb')
    int_array = array.array('i')
    float_array = array.array('f')
    int_array.fromfile(f, 2)
    irec0, Nw=int_array[-2:]
    float_array.fromfile(f, 3)
    wl_min, dwl, FWHM = float_array[-3:]
    int_array.fromfile(f, 5)
    Np, Nt, Ng, mol_id, isotopolog_id = int_array[-5:]
    try:
        mol_name=list(nemesis_hitran_id_numbers.keys())[ \
            list(nemesis_hitran_id_numbers.values()).index(mol_id)]
    except:
        mol_name=None
    float_array.fromfile(f, Ng)
    ggrid = np.array(float_array[-Ng:])
    float_array.fromfile(f, Ng)
    weights = np.array(float_array[-Ng:])
    float_array.fromfile(f, 2)
    float_array.fromfile(f, Np)
    pgrid = np.array(float_array[-Np:])
    float_array.fromfile(f, Nt)
    tgrid = np.array(float_array[-Nt:])
    ntot=Nw*Np*Nt*Ng
    if dwl>=0.: #regular grid
        wls=wl_min+np.arange(Nw)*dwl
    else:
        float_array.fromfile(f, Nw)
        wls = np.array(float_array[-Nw:])
    wns=10000./wls[::-1]
    f.close()
    f = open(filename, mode='rb') # restart to start reading kdata at record number irec0
    kdata=array.array('f')
    kdata.fromfile(f, irec0-1+ntot)
    kdata=np.reshape(kdata[-ntot:],(Nw,Np,Nt,Ng))[::-1]*1.e-20
    kdata=kdata.transpose(1,2,0,3)
    f.close()
    return Np, Nt, Nw, Ng, pgrid, tgrid, wns, ggrid, weights, \
        kdata, mol_name, isotopolog_id



# -*- coding: utf-8 -*-
"""
@author: jeremy leconte
"""
from datetime import date
import numpy as np
import h5py
import pkg_resources  # part of setuptools
import astropy.units as u
from .util.interp import rm_molec,unit_convert,interp_ind_weights,bilinear_interpolation
from .settings import Settings
from .util.spectral_object import Spectral_object
from .util.molar_mass import Molar_mass
from .util.cst import ktable_long_name_attributes

class Data_table(Spectral_object):
    """An abstract class that will serve as a basis for Ktable and Xtable.
    This class includes all the interpolation and remapping methods.
    """

    __array_priority__=1000
    # this allows __rmul__ to take precedence over numpy array broadcasting in multiplications
    # See https://stackoverflow.com/questions/40694380/
    # forcing-multiplication-to-use-rmul-instead-of-numpy-array-mul-or-byp/44634634#44634634

    def __init__(self):
        """Initializes all attributes to `None`"""
        self.filename=None
        self.mol=None
        self.isotopolog_id=0
        self.pgrid=None
        self.logpgrid=None
        self.tgrid=None
        self.wns=None
        self.wnedges=None
        self.kdata=None
        self.logk=None
        self.Np=None
        self.Nt=None
        self.Nw=None
        self.Ng=None   # If the table is for xsec, Ng will stay at None.
                       # This will allow us to differentiate xsec from corrk
                       # when needed in the interpolation routines.
                       # Especially to reshape the output.
        self.DOI='unknown'
        self.sampling_method='unknown'
        version = pkg_resources.require("exo_k")[0].version
        self.Date_ID='exo_k-v'+version+'-'+date.today().strftime("%d/%m/%Y")
        self.Nx=None   # If we are dealing with a Qtable with a variable gas, Nx is the size of the
                       # grid along the dimension of the Volume mixing ratio of the variable gas.
                       # A None value means that we have either a regular Ktable or a Xtable.
        self.p_unit='unspecified'
        self.kdata_unit='unspecified'
        self.wn_unit='cm^-1'
        self._settings=Settings()

    def finalize_init(self, p_unit='unspecified', file_p_unit='unspecified',
        kdata_unit='unspecified', file_kdata_unit='unspecified',
        remove_zeros=False):
        """Common code at the end of the initialization of
        inheriting classes put here to avoid duplicates
        """
        if self.kdata is not None:
            if self._settings._convert_to_mks:
                if p_unit is 'unspecified': p_unit='Pa'
                if kdata_unit is 'unspecified': kdata_unit='m^2/molecule'
            self.convert_p_unit(p_unit=p_unit,file_p_unit=file_p_unit)
            self.convert_kdata_unit(kdata_unit=kdata_unit,file_kdata_unit=file_kdata_unit)
            if remove_zeros : self.remove_zeros(deltalog_min_value=10.)

    def copy_attr(self, other, cp_kdata=False):
        """Copy attributes from other

        Parameters
        ----------
            other: :class:`Data_table`
                :class:`Data_table` object that will be copied
            cp_kdata: bool, optional
                If `False`, only metadata are copied
        """
        self.filename=other.filename
        self.mol     =other.mol
        self.pgrid   =np.copy(other.pgrid)
        self.logpgrid=np.copy(other.logpgrid)
        self.tgrid   =np.copy(other.tgrid)
        self.wns     =np.copy(other.wns)
        self.wnedges =np.copy(other.wnedges)
        self.Np      =other.Np
        self.Nt      =other.Nt
        self.Nw      =other.Nw
        self.Ng      =other.Ng
        self.Nx      =other.Nx
        self.logk    =other.logk
        self.p_unit  =other.p_unit
        self.kdata_unit= other.kdata_unit
        if cp_kdata:
          self.kdata=np.copy(other.kdata)
        else:
          self.kdata=None

    def remove_zeros(self,deltalog_min_value=10.):
        """Finds zeros in the kdata and set them to (10.**-deltalog_min_value)
        times the minimum positive value in the table (inplace).

        This is to be able to work in logspace. 

        Parameters
        ----------
            deltalog_min_value: float        
        """
        mask = np.zeros(self.kdata.shape,dtype=bool)
        mask[np.nonzero(self.kdata)] = True
        minvalue=np.amin(self.kdata[mask])
        self.kdata[~mask]=minvalue/(10.**deltalog_min_value)

    def convert_p_unit(self, p_unit='unspecified', file_p_unit='unspecified'):
        """Converts pressure to a new unit (inplace).

        Parameters
        ----------
            p_unit: str
                String identifying the pressure units to convert to (e.g. 'bar', 'Pa', 'mbar', 
                or any pressure unit recognized by the astropy.units library).
                If ='unspecified', no conversion is done.
            file_p_unit : str, optional
                String to specify the current pressure unit if it is unspecified or if 
                you have reasons to believe it is wrong (e.g. you just read a file where
                you know that the pressure grid and the pressure unit do not correspond)
        """
        #if p_unit==file_p_unit and self.p_unit != 'unspecified': return
        current_p_unit=self.p_unit
        self.p_unit,conversion_factor=unit_convert( \
            'p_unit',unit_file=current_p_unit,unit_in=file_p_unit,unit_out=p_unit)
        self.pgrid=self.pgrid*conversion_factor
        self.logpgrid=np.log10(self.pgrid)

    def convert_kdata_unit(self, kdata_unit='unspecified', file_kdata_unit='unspecified'):
        """Converts kdata to a new unit (inplace).

        Parameters
        ----------
            kdata_unit: str
                String to identify the units to convert to. Accepts 'cm^2', 'm^2'
                or any surface unit recognized by the astropy.units library.
                If ='unspecified', no conversion is done.
                In general, kdata should be kept in 'per number' or 'per volume'
                units (as opposed to 'per mass' units) as composition will
                always be assumed to be a number or volume mixing ratio.
                Opacities per unit mass are not supported yet.
                Note that you do not need to specify the '/molec' or
                '/molecule' in the unit.
            file_kdata_unit : str
                String to specify the current kdata unit if it is unspecified or if 
                you have reasons to believe it is wrong (e.g. you just read a file where
                you know that the kdata grid and the kdata unit do not correspond)
        """
        #if kdata_unit==file_kdata_unit and self.kdata_unit != 'unspecified': return
        tmp_k_u_in=rm_molec(file_kdata_unit)
        tmp_k_u_out=rm_molec(kdata_unit)
        tmp_k_u_file=rm_molec(self.kdata_unit)
        self.kdata_unit,conversion_factor=unit_convert(  \
            'kdata_unit',unit_file=tmp_k_u_file,unit_in=tmp_k_u_in,unit_out=tmp_k_u_out)
        self.kdata_unit=self.kdata_unit+'/molecule'
        self.kdata=self.kdata*conversion_factor
    
    def convert_to_mks(self):
        """Converts units to MKS (inplace).
        """
        self.convert_kdata_unit(kdata_unit='m^2')
        self.convert_p_unit(p_unit='Pa')

    def interpolate_kdata(self, logp_array=None, t_array=None, x_array=1.,
            log_interp=None, logp_interp=True, wngrid_limit=None):
        """interpolate_kdata interpolates the kdata at on a given temperature and
        log pressure profile. If a volume mixing ratio profile (`x_array`) is given,
        the cross section computed for the species is multiplied by `x_array` to account for the
        'dilution' of the opacity.

        Parameters
        ----------
            logp_array: array
                log 10 pressure array to interpolate to
            t_array: array, same size a logp_array
                Temperature array to interpolate to
            x_array: None
                Volume mixing ratio array used to renormalize the cross section.

        If floats are given, they are interpreted as arrays of size 1.

        Other Parameters
        ----------------
            wngrid_limit: list or array
                if an array of to values is given, interpolates only within this array
            log_interp: bool, optional
                Whether the interpolation is linear in kdata or in log(kdata)

        Returns
        -------
            array of shape (logp_array.size, self.Nw (, self.Ng))
                The interpolated kdata.

        """
        if hasattr(logp_array, "__len__"):
            logp_array=np.array(logp_array, dtype=float)
        else:
            logp_array=np.array([logp_array], dtype=float)
        if hasattr(t_array, "__len__"):
            t_array=np.array(t_array)
        else:
            t_array=np.array([t_array])
        if hasattr(x_array, "__len__"):
            x_array=np.array(x_array)
        else:
            x_array=np.array([x_array])
        tind,tweight=interp_ind_weights(t_array,self.tgrid)
        if logp_interp:
            lpind,lpweight=interp_ind_weights(logp_array,self.logpgrid)
        else:
            lpind,lpweight=interp_ind_weights(10**logp_array,self.pgrid)
        if wngrid_limit is None:
            wngrid_filter = slice(None)
            Nw=self.Nw
        else:
            wngrid_limit=np.array(wngrid_limit)
            wngrid_filter = np.where((self.wnedges > wngrid_limit.min()) & (
                self.wnedges <= wngrid_limit.max()))[0][:-1]
            Nw=wngrid_filter.size
        if self.Ng is None:
            res=np.zeros((tind.size,Nw))
        else:
            res=np.zeros((tind.size,Nw,self.Ng))
        if log_interp is None: log_interp=self._settings._log_interp
        if log_interp:
            for ii in range(tind.size):
                kc_p1t1=np.log(self.kdata[lpind[ii],tind[ii]][wngrid_filter].ravel())
                kc_p0t1=np.log(self.kdata[lpind[ii]-1,tind[ii]][wngrid_filter].ravel())
                kc_p1t0=np.log(self.kdata[lpind[ii],tind[ii]-1][wngrid_filter].ravel())
                kc_p0t0=np.log(self.kdata[lpind[ii]-1,tind[ii]-1][wngrid_filter].ravel())
                res[ii]=np.reshape(bilinear_interpolation(kc_p0t0, kc_p1t0, 
                    kc_p0t1, kc_p1t1, lpweight[ii], tweight[ii]),(Nw,-1)).squeeze()
            return (x_array*np.exp(res).transpose()).transpose()
            # trick for the broadcasting to work whatever the shape of x_array
        else:
            for ii in range(tind.size):
                kc_p1t1=self.kdata[lpind[ii],tind[ii]][wngrid_filter].ravel()
                kc_p0t1=self.kdata[lpind[ii]-1,tind[ii]][wngrid_filter].ravel()
                kc_p1t0=self.kdata[lpind[ii],tind[ii]-1][wngrid_filter].ravel()
                kc_p0t0=self.kdata[lpind[ii]-1,tind[ii]-1][wngrid_filter].ravel()
                res[ii]=np.reshape(bilinear_interpolation(kc_p0t0, kc_p1t0, 
                    kc_p0t1, kc_p1t1, lpweight[ii], tweight[ii]),(Nw,-1)).squeeze()
            return (x_array*res.transpose()).transpose()
            # trick for the broadcasting to work whatever the shape of x_array

    def remap_logPT(self, logp_array=None, t_array=None, x_array=None):
        """remap_logPT re-interpolates the kdata on a new temperature and log pressure grid
        (inplace). 

        Parameters
        ----------
            logp_array: Array
                log 10 pressure array to interpolate to
            t_array: Array
                temperature array to interpolate to
            x_array: dummy argument to be consistent with interpolate_kdata in Ktable5d

        Whether the interpolation is linear in kdata or in log10(kdata)
        is controlled by self._settings._log_interp
        """
        if x_array is not None: print('be careful, providing an x_array is usually for Ktable5d')
        t_array=np.array(t_array)
        logp_array=np.array(logp_array, dtype=float)
        tind,tweight=interp_ind_weights(t_array,self.tgrid)
        lpind,lpweight=interp_ind_weights(logp_array,self.logpgrid)
        lpindextended=lpind[:,None]
        if self.Ng is None:
            tw=tweight[None,:,None]    # trick to broadcast over Nw and Ng a few lines below
            pw=lpweight[:,None,None]
        else:
            tw=tweight[None,:,None,None]    # trick to broadcast over Nw and Ng a few lines below
            pw=lpweight[:,None,None,None]
        #tw=tweight.reshape((1,tweight.size,1,1))  
        ## trick to broadcast over Nw and Ng a few lines below
        #pw=lpweight.reshape((lpweight.size,1,1,1))
        kc_p1t1=self.kdata[lpindextended,tind]
        kc_p0t1=self.kdata[lpindextended-1,tind]
        kc_p1t0=self.kdata[lpindextended,tind-1]
        kc_p0t0=self.kdata[lpindextended-1,tind-1]
        if self._settings._log_interp is True:
            kdata_tmp=  np.log10(kc_p1t1)*pw*tw          \
                            +np.log10(kc_p0t1)*(1.-pw)*tw \
                            +np.log10(kc_p1t0)*pw*(1.-tw) \
                            +np.log10(kc_p0t0)*(1.-pw)*(1.-tw)
            self.kdata=10**kdata_tmp
        else:
            kdata_tmp=  (kc_p1t1)*pw*tw          \
                        +(kc_p0t1)*(1.-pw)*tw \
                        +(kc_p1t0)*pw*(1.-tw) \
                        +(kc_p0t0)*(1.-pw)*(1.-tw)
            self.kdata=kdata_tmp
        self.logpgrid=logp_array
        self.pgrid   =10**self.logpgrid
        self.tgrid   =t_array
        self.Np      =logp_array.size
        self.Nt      =t_array.size

    def pindex(self, p):
        """Finds and returns the index corresponding to the given pressure p
        (units must be the same as the ktable)
        """
        return min(np.searchsorted(self.pgrid,p),self.Np-1)

    def tindex(self, t):
        """Finds and returns the index corresponding to the given temperature t (in K)
        """
        return min(np.searchsorted(self.tgrid,t),self.Nt-1)

    def wlindex(self, wl):
        """Finds and returns the index corresponding to the given wavelength (in microns)
        """
        return min(np.searchsorted(self.wns,10000./wl),self.Nw-1)-1
        #return min(np.searchsorted(self.wnedges,10000./wl),self.Nw-1)-1

    def __repr__(self):
        """Method to output header
        """
        output="""
        file         : {file}
        molecule     : {mol}
        p grid       : {p}
        p unit       : {p_unit}
        t grid   (K) : {t}
        wn grid      : {wn}
        wn unit      : {wnu}
        kdata unit   : {kdata_unit}""".format(file=self.filename,mol=self.mol,
            p=self.pgrid, p_unit=self.p_unit, t=self.tgrid,
            wn=self.wns, wnu=self.wn_unit,
            kdata_unit=self.kdata_unit)
        return output

    def plot_spectrum(self, ax, p=1.e-5, t=200., x=1., g=None,
            x_axis='wls', xscale=None, yscale=None, **kwarg):
        """Plot the spectrum for a given point

        Parameters
        ----------
            ax : :class:`pyplot.Axes`
                A pyplot axes instance where to put the plot.
            p : float
                Pressure (Ktable pressure unit)
            t : float
                Temperature (K)
            g: float
                Gauss point
            x: float
                Mixing ratio of the species

        Other Parameters
        ----------------
            x_axis: str, optional
                If 'wls', x axis is wavelength. Wavenumber otherwise.
            x/yscale: str, optional
                If 'log' log axes are used.
        """
        toplot=self.spectrum_to_plot(p=p, t=t, x=x, g=g)
        if x_axis == 'wls':
            ax.plot(self.wls,toplot,**kwarg)
            ax.set_xlabel('Wavelength (micron)')
        else:
            ax.plot(self.wns,toplot,**kwarg)
            ax.set_xlabel('Wavenumber (cm$^{-1}$)')
        ax.set_ylabel('Cross section ('+self.kdata_unit+')')
        if xscale is not None: ax.set_xscale(xscale)
        if yscale is not None: ax.set_yscale(yscale)

    def spectrum_to_plot(self, p=1.e-5, t=200., x=1., g=None):
        """Dummy function to be defined in inheriting classes
        """
        raise NotImplementedError()

    def vmr_normalize(self, x_self):
        """Rescales kdata to account for the fact that the gas is not a pure species

        Parameters
        ----------
            x_self: float or array of shape (`self.Np,self.Nt`)
                The volume mixing ratio of the species.
        
        Returns
        -------
            array
                The vmr normalized kdata (x_self*self.kdata).
        """
        if x_self is None : return self.kdata
        if isinstance(x_self, (float,int)): return x_self*self.kdata
        if not isinstance(x_self,np.ndarray):
            print("""in vmr_normalize:
            x_self should be a float or a numpy array: I'll probably stop now!""")
            raise TypeError('bad mixing ratio type')
        if np.array_equal(x_self.shape,self.kdata.shape[0:2]):
            if self.Ng is None:
                return x_self[:,:,None]*self.kdata
            else:                
                return x_self[:,:,None,None]*self.kdata
        else:
            print("""in vmr_normalize:
            x_self shape should be (pgrid.size,tgrid.size): I'll stop now!""")
            raise TypeError('bad mixing ratio type')

    def __rmul__(self, vmr):
        """Defines the right "*" operator with a vmr
        """
        res=self.copy()
        res.kdata=self.vmr_normalize(vmr)
        return res

    __mul__ = __rmul__

    def combine_with(self, other, x_self=None, x_other=None, **kwargs):
        """Method to create a new `Data_table` where the kdata of 'self' are:

            * randomly mixed with 'other' for a `Ktable`.
            * added to 'other' for an `Xtable`

        Parameters
        ----------
            other : same class as self
                A :class:`Data_table` object to be mixed with.
                Dimensions should be the same as self.
            x_self : float or array, optional
                Volume mixing ratio of self.
            x_other : float or array, optional
                Volume mixing ratio of the species to be mixed with (other).

        If either x_self or x_other are set to `None` (default),
        the cross section of the species in question
        are considered to be already normalized with respect to the mixing ratio.

        Returns
        -------
            :class:`Data_table`
                A new table for the mix
        """
        if other.Nx is not None:
            # if other is a Ktable5d, use the method for this class instead. 
            return other.combine_with(self, x_self=x_other, x_other=x_self, **kwargs)
        if not np.array_equal(self.shape,other.shape):
            raise TypeError("""in combine_with: kdata tables do not have the same dimensions.
                I'll stop now!""")
        if (self.p_unit!=other.p_unit) or \
            (rm_molec(self.kdata_unit)!=rm_molec(other.kdata_unit)):
            raise RuntimeError("""in combine_with: tables do not have the same units.
                I'll stop now!""")
        res=self.copy(cp_kdata=False)
        if self.Ng is None:
            res.kdata=self.vmr_normalize(x_self) + other.vmr_normalize(x_other)
        else:
            res.kdata=self.RandOverlap(other, x_self, x_other, **kwargs)
        return res

    def __add__(self, other):
        """Defines the "+" operator with another Data_table

        .. warning::
            __radd__ is not implemented because we want to use the __add__ (or combine_with)
            method of the left object.
        """
        return self.combine_with(other)

    def __getitem__(self, key):
        """To access the data without typing self.kdata[]

        Parameters
        ----------
            key: can be slices, like for a numpy array.
        """
        return self.kdata[key]

    def clip_spectral_range(self, wn_range=None, wl_range=None):
        """Limits the data to the provided spectral range (inplace):

           * Wavenumber in cm^-1 if using wn_range argument
           * Wavelength in micron if using wl_range
        """
        if (wn_range is None) and (wl_range is None): return
        if wl_range is not None:
            if wn_range is not None:
                raise RuntimeError('Should provide either wn_range or wl_range, not both!')
            _wn_range=np.sort(10000./np.array(wl_range))
        else:
            _wn_range=np.sort(np.array(wn_range))
        iw_min, iw_max=np.searchsorted(self.wnedges, _wn_range, side='left')
        iw_max-=1
        self.wnedges=self.wnedges[iw_min:iw_max+1]
        self.wns=self.wns[iw_min:iw_max]
        self.Nw=self.wns.size
        if self.Nx is None:
            self.kdata=self.kdata[:,:,iw_min:iw_max]
        else:
            self.kdata=self.kdata[:,:,:,iw_min:iw_max]

    def extend_spectral_range(self, wngrid_left=None, wngrid_right=None,
            wnedges_left=None, wnedges_right=None,
            remove_zeros=False):
        """Extends the spectral range of an existing table (inplace).
        The new bins are filled with zeros (except if remove_zeros=True)

        Parameters
        ----------
            wngrid_left: array
                Array of wavenumbers to add to the small wn end of the table.
            wngrid_right: array
                Array of wavenumbers to add to the high wn end of the table.

        .. warning::
            There should not be any overlap between wngrid_left, wngrid_right,
            and the current wavenumber grid of the table. 

        Other Parameters
        ----------------
            remove_zeros: bool (optional)
                Whether zeros in the resulting table should be removed using
                :func:`remove_zeros`.
        """
        new_wns=[]
        new_wnedges=[]
        idx_offset=0
        if wngrid_left is not None:
            wngrid_left=np.array(wngrid_left, dtype=float)
            if wngrid_left[-1]>=self.wnedges[0]:
                raise RuntimeError("The left grid overlaps with the current one")
            new_wns.append(wngrid_left)
            if wnedges_left is not None:
                wnedges_left=np.array(wnedges_left, dtype=float)
                new_wnedges.append(wnedges_left)
            else:
                new_wnedges.append([wngrid_left[0]])
                new_wnedges.append(0.5*(wngrid_left[:-1]+wngrid_left[1:]))
            idx_offset=wngrid_left.size
        else:
            if wnedges_left is not None:
                if wnedges_left[-1]>=self.wnedges[0]:
                    raise RuntimeError("The left grid overlaps with the current one")
                wnedges_left=np.array(wnedges_left, dtype=float)
                new_wnedges.append(wnedges_left)
                new_wns.append(0.5*(wnedges_left[:-1]+wnedges_left[1:]))
                new_wns.append([0.5*(wnedges_left[-1]+self.wnedges[0])])
                idx_offset=wnedges_left.size

        new_wns.append(self.wns)
        new_wnedges.append(self.wnedges)
        if wngrid_right is not None:
            wngrid_right=np.array(wngrid_right, dtype=float)
            if wngrid_right[0]<=self.wnedges[-1]:
                raise RuntimeError("The right grid overlaps with the current one")
            new_wns.append(wngrid_right)
            if wnedges_right is not None:
                wnedges_right=np.array(wnedges_right, dtype=float)
                new_wnedges.append(wnedges_right)
            else:
                new_wnedges.append(0.5*(wngrid_right[:-1]+wngrid_right[1:]))
                new_wnedges.append([wngrid_right[-1]])
        else:
            if wnedges_right is not None:
                wnedges_right=np.array(wnedges_right, dtype=float)
                if wnedges_right[0]<=self.wnedges[-1]:
                    raise RuntimeError("The right grid overlaps with the current one")
                new_wnedges.append(wnedges_right)
                new_wns.append([0.5*(wnedges_right[0]+self.wnedges[-1])])
                new_wns.append(0.5*(wnedges_right[:-1]+wnedges_right[1:]))
        new_wns=np.concatenate(new_wns)
        new_wnedges=np.concatenate(new_wnedges)
        newshape=self.shape
        if self.Nx is None:
            newshape[2]=new_wns.size
            newkdata=np.zeros(newshape)
            newkdata[:,:,idx_offset:idx_offset+self.Nw]=self.kdata
        else:
            newshape[3]=new_wns.size
            newkdata=np.zeros(newshape)
            newkdata[:,:,:,idx_offset:idx_offset+self.Nw]=self.kdata
        self.kdata=newkdata
        self.wns=new_wns
        self.wnedges=new_wnedges
        self.Nw=self.wns.size
        if remove_zeros : self.remove_zeros(deltalog_min_value=10.)

    def bin_down_cp(self, wnedges=None, **kwargs):
        """Creates a copy of the instance and bins it down using the methods in 
        Ktable or Xtable.

        See :func:`exo_k.ktable.Ktable.bin_down` or :func:`exo_k.xtable.Xtable.bin_down`
        for details on parameters.

        Returns
        -------
            :class:`~exo_k.ktable.Ktable` or :class:`~exo_k.xtable.Xtable` object
                The binned down table.
        """
        res=self.copy()
        res.bin_down(wnedges=wnedges, **kwargs)
        return res

    def change_molecule_name(self, mol_name):
        """Changes name of the molecule (self.mol attribute).

        Parameters
        ----------
            mol_name: str
                New molecule name
        """
        self.mol=mol_name

    @property
    def molar_mass(self):
        """Computes molar mass from molecule name
        """
        return Molar_mass().fetch(self.mol)

    def write_hdf5_common(self, f, compression="gzip", compression_level=9,
        p_unit=None):
        """Method that writes datasets and attributes that are common to
        X and Ktables. 

        Parameters
        ----------
            f: h5py file instance
                The file to write that is created in daughter classes
        """
        dt = h5py.string_dtype(encoding='utf-8')
        f.create_dataset("DOI", (1,), data=self.DOI, dtype=dt)
        f.create_dataset("Date_ID", (1,), data=self.Date_ID, dtype=dt)
        f.create_dataset("mol_name", (1,), data=self.mol, dtype=dt)
        f.create_dataset("t", data=self.tgrid,
            compression=compression, compression_opts=compression_level)
        f["t"].attrs["units"] = 'K'

        if p_unit is not None:
            conv_factor=u.Unit(self.p_unit).to(u.Unit(p_unit))
            data_to_write=self.pgrid*conv_factor
            f.create_dataset("p", data=data_to_write,
                compression=compression, compression_opts=compression_level)
            f["p"].attrs["units"] = p_unit
        else:
            f.create_dataset("p", data=self.pgrid,
                compression=compression, compression_opts=compression_level)
            f["p"].attrs["units"] = self.p_unit

        f.create_dataset("wnrange", data=self.wnrange,
            compression=compression, compression_opts=compression_level)
        f.create_dataset("wlrange", data=self.wlrange,
            compression=compression, compression_opts=compression_level)

        for key,val in ktable_long_name_attributes.items():
            if key in f: f[key].attrs["long_name"] = val
        for key in ('bin_edges', 'bin_centers', 'wnrange'):
            if key in f: f[key].attrs["units"] = self.wn_unit
        if 'wlrange' in f: f['wlrange'].attrs["units"] = 'micron'


    def toLogK(self):
        """Changes kdata to log 10.
        """
        if not self.logk:
            self.logk=True
            self.kdata=np.log10(self.kdata)
        return
            
    def toLinK(self):
        """Changes kdata back from log to linear scale.
        """
        if self.logk:
            self.logk=False
            self.kdata=np.power(10.,self.kdata)
        return



"""Defines constants"""

G = 6.67384e-11
RSOL = 6.955e8
RJUP = 7.1492e7
RJUP_POL = 6.9911e7 # Polar Radius
PI = 3.14159265359
MSOL = 1.98847542e+30
MJUP = 1.898e27
AU = 1.49597871e+11
KBOLTZ = 1.380648813e-23
RGP = 8.31446
PLANCK = 6.62606957e-34
C_LUM = 299792458
SIG_SB = 5.670367e-08
N_A = 6.022140857e+23
tiny=1.e-13
OneMintiny=1.-tiny

ktable_long_name_attributes={
    'p': 'Pressure grid',
    't': 'Temperature grid',
    'x': 'Volume mixing ratio grid',
    'bin_centers': 'Centers of the wavenumber bins',
    'bin_edges': 'Separations between the wavenumber bins',
    'weights': 'Weights used in the g-space quadrature',
    'samples': 'Abscissas used to sample the k-coefficients in g-space',
    'mol_name': 'Name of the species described',
    'mol_mass': 'Mass of the species',
    'kcoeff': 'Table of the k-coefficients with axes (pressure, temperature, wavenumber, g space)',
    'xsecarr': 'Table of cross sections with axes (pressure, temperature, wavenumber)',
    'method': 'Name of the method used to sample g-space',
    'ngauss': 'Number of points used to sample the g-space',
    'temperature_grid_type': 'Whether the temperature grid is "regular" (same temperatures for all pressures) or "pressure-dependent"',
    'wnrange': 'Wavenumber range covered',
    'wlrange': 'Wavelength range covered',
    'DOI': 'Data object identifier linked to the data',
    'Date_ID':'Date at which the table has been created, along with the version of exo_k',
    'key_iso_ll': 'Isotopologue identifier'
}

nemesis_hitran_id_numbers={
'H2O': 1,
'CO': 5,
'CO2': 2,
'O2': 7,
'O3': 3,
'N2O': 4,
'AlCl': 84,
'AlF': 85,
'Al': 86,
'AlO': 83,
'AsH3': 41,
'BeH': 87,
'C2': 88,
'C2H2': 26,
'C2H4': 32,
'CaF': 89,
'CaH': 90,
'CaO': 92,
'CH': 93,
'CH3': 94,
'CH3Cl': 24,
'CH3F': 95,
'CH4': 6,
'CN': 96,
'CP': 97,
'CrH': 98,
'CS': 54,
'FeH': 82,
'H2': 39,
'H2CO': 20,
'H2O2': 25,
'H2S': 36,
'H3^+': 81,
'HBr': 16,
'HCl': 15,
'HCN': 23,
'HD+': 99,
'HeH+': 100,
'HF': 14,
'HI': 17,
'HNO3': 12,
'K': 61,
'KCl': 101,
'KF': 102,
'LiCl': 103,
'LiF': 104,
'LiH': 105,
'LiH+': 106,
'MgF': 107,
'MgH': 108,
'MgO': 109,
'Na': 60,
'NaCl': 110,
'NaF': 111,
'NaH': 112,
'NH': 113,
'NH3': 11,
'NO': 8,
'NO2': 10,
'NS': 114,
'OH': 13,
'OH+': 115,
'P2H2': 116,
'PH': 118,
'PH3': 28,
'PN': 119,
'PO': 120,
'PS': 121,
'ScH': 122,
'SH': 123,
'SiH': 124,
'SiH2': 125,
'SiH4': 126,
'SiO': 127,
'SiO2': 128,
'SiS': 129,
'SO2': 9,
'SO3': 74,
'TiH': 130,
'TiO': 62,
'VO': 63
}



# -*- coding: utf-8 -*-
"""
@author: jeremy leconte

Library of useful functions for handling filenames
"""
import numpy as np
import h5py

class EndOfFile(Exception):
    """Error for an end of file
    """

def select_kwargs(kwargs, filter_keys_list=[]):
    """Function to select only some keyword arguments from a
    kwargs dict to pass to a function afterward.
    
    Parameters
    ----------
        kwargs: dict
            Dictionary of keyword arguments and values.
        filter_keys_list: list of str
            Names of the keys to select.

    Returns
    -------
        filtered_kwargs: dict
            A dictionary with only the selected keys (if they were
            present in kwargs).
        
    Examples
    --------
        >>> def func(allo=None):
        >>>     print(allo)
        >>>
        >>> def bad_global_func(**kwargs):
        >>>     print(kwargs)
        >>>     func(**kwargs)
        >>>
        >>> def good_global_func(**kwargs):
        >>>     print(kwargs)
        >>>     func(**select_kwargs(kwargs,['allo']))

        >>> bad_global_func(allo=3.,yeah=None)        
        {'allo': 3.0, 'yeah': None}
        TypeError: func() got an unexpected keyword argument 'yeah'

        >>> good_global_func(allo=3.,yeah=None)        
        {'allo': 3.0, 'yeah': None}
        3.0

    """
    filtered_kwargs=dict()
    for key in filter_keys_list:
        if key in kwargs.keys():
            filtered_kwargs[key]=kwargs[key]
    return filtered_kwargs

def create_fname_grid_Kspectrum_LMDZ(Np, Nt, Nx=None, suffix='', nb_digit=3):
    """Creates a grid of filenames consistent with Kspectrum/LMDZ
    from the number of pressure, temperatures (, and vmr) points (respectively) in the grid.

    Parameters
    ----------
        Np, Nt: int
            Number of Pressure and temperature points.
        Nx: int, optional
            Number of vmr points if there is a variable gas
        suffix: str, optional
            String to add behind the 'k001'
        nb_digit: int, optional
            Total number of digit including the zeros ('k001' is 3)
    
    Returns
    -------
        list of str
            List of the files in the right order and formating to be 
            given to :func:`exo_k.ktable.Ktable.hires_to_ktable` or 
            :func:`exo_k.xtable.Xtable.hires_to_xtable`.

    Examples
    --------

        >>> exo_k.create_fname_grid_Kspectrum_LMDZ(2,3)                  
        array([['k001', 'k002', 'k003'],
        ['k004', 'k005', 'k006']], dtype='<U4')

        >>> exo_k.create_fname_grid_Kspectrum_LMDZ(2,3,suffix='.h5')                  
        array([['k001.h5', 'k002.h5', 'k003.h5'],
        ['k004.h5', 'k005.h5', 'k006.h5']], dtype='<U7')

    """    
    res=[]
    ii=1
    if Nx is None:
        for _ in range(Np):
            for _ in range(Nt):
                res.append('k'+str(ii).zfill(nb_digit)+suffix)
                ii+=1
        return np.array(res).reshape((Np,Nt))
    else:
        for _ in range(Nx):
            for _ in range(Np):
                for _ in range(Nt):
                    res.append('k'+str(ii).zfill(nb_digit)+suffix)
                    ii+=1
        res=np.array(res).reshape((Nx,Np,Nt))
        return np.transpose(res,(2,1,0))

def convert_exo_transmit_to_hdf5(file_in, file_out, mol='unspecified'):
    """Converts exo_transmit like spectra to hdf5 format for speed and space.

    Parameters
    ----------
        file_in: str
            Initial exo_transmit filename.
        file_out: str
            Name of the final hdf5 file to be created.
    """
    tmp_wlgrid=[]
    tmp_kdata=[]
    with open(file_in, 'r') as file:
        tmp = file.readline().split()
        tgrid=np.array([float(ii) for ii in tmp])
        tmp = file.readline().split()
        pgrid=np.array([float(ii) for ii in tmp])
        Np=pgrid.size
        Nt=tgrid.size
        while True:
            line=file.readline()
            if line is None or line=='': break
            tmp_wlgrid.append(float(line.split()[0])) # wavelength in m
            tmp_kdata.append([])
            for _ in range(Np):
                tmp_kdata[-1].append(file.readline().split()[1:])
    tmp_wlgrid=np.array(tmp_wlgrid)
    Nw=tmp_wlgrid.size
    kdata=np.zeros((Np,Nt,Nw))
    for iP in range(Np):
        for iT in range(Nt):
            for iW in range(Nw):
                kdata[iP,iT,iW]=tmp_kdata[Nw-iW-1][iP][iT]
    print(kdata.shape)
    wns=0.01/tmp_wlgrid[::-1]
    
    if not file_out.lower().endswith(('.hdf5', '.h5')):
        fullfilename=file_out+'.h5'
    else:
        fullfilename=file_out
    compression="gzip"
    f = h5py.File(fullfilename, 'w')
    f.attrs["mol_name"] = mol
    f.create_dataset("p", data=pgrid,compression=compression)
    f["p"].attrs["units"] = 'Pa'
    f.create_dataset("t", data=tgrid,compression=compression)
    f.create_dataset("xsecarr", data=kdata,compression=compression)
    f["xsecarr"].attrs["units"] = 'm^2'
    f.create_dataset("bin_edges", data=wns,compression=compression)
    f.close()    

def _read_array(file, Nvalue, N_per_line=5, Nline=None, revert=False):
    """Reads an array in a .dat filestream. 
    Assumes that the arrays are arranged N_per_line values per line.

    Parameters
    ----------
        file: file stream
            File to be read.
        Nvalue: int
            Number of values to be read. 
        N_per_line: int
            Number of values assumed per line.
        revert: boolean (optional)
            Whether or not to revert the array (because some arrays are
            in decreasing order).
        Nline: int (optional)
            Slightly speeds up the process if the user specifies
            the number of lines that will need to be read. 

    Returns
    -------
        Array
            A numpy array with the values.
    """
    
    if Nline is None:
        Nline=Nvalue//N_per_line
        if Nvalue%N_per_line != 0:
            Nline+=1
    new_array=[]
    for _ in range(Nline):
        line=[float(i) for i in file.readline().split()]
        new_array+=line
    if revert :
        new_array=np.array(new_array)[::-1]
    else:
        new_array=np.array(new_array)
    #print(new_array)
    return new_array

def _read_exorem_k_array(file, Nline):
    """Reads an array in a .dat filestream. 
    Assumes that the arrays are arranged N_per_line values per line.

    Parameters
    ----------
        file: file stream
            File to be read.
        Nline: int
            Number of values/lines to be read. 

    Returns
    -------
        Array
            A numpy array with the values.
    """
    
    new_array=[]
    for _ in range(Nline):
        srt=file.readline()
        try:
            new_array.append(float(srt))
        except ValueError:
            i = srt.index('-')
            srt = srt[:i] + 'e' + srt[i:]
            new_array.append(float(srt))
    new_array=np.array(new_array)
    return new_array



####### OLD FUNCTIONS ###################
def convert_old_kspectrum_to_hdf5(file_in, file_out, skiprows=0):
    """Deprecated.
    
    Converts kspectrum like spectra to hdf5 format for speed and space

    Parameters
    ----------
        file_in: str
            Initial kspectrum filename.
        file_out: str
            Name of the final hdf5 file to be created.
    """
    wn_hr,k_hr=np.loadtxt(file_in,skiprows=skiprows,unpack=True) 
    f = h5py.File(file_out, 'w')
    f.create_dataset("wns", data=wn_hr,compression="gzip")
    f.create_dataset("k", data=k_hr,compression="gzip")
    f.close()   



# -*- coding: utf-8 -*-
"""
@author: jeremy leconte

A class based on a singleton to store global options only once for every instance. 
"""
import os.path
import re
from glob import glob
from string import Template
from .util.singleton import Singleton

class Settings(Singleton):
    """A class based on a singleton to store global options only once for every instance.

    So all the following methods can be called using the following syntax:

    >>> exo_k.Settings().method_name(agrs)

    In gerneal, they will change internal global attributes
    that change the global behavior of some routines in the library. 
    """

    def init(self, *args, **kwds):
        self.reset_search_path()
        self._log_interp = True
        self._convert_to_mks = False
        self._delimiter = '_'
        self.set_delimiters(r"_.\-")

        self._case_sensitive = False

    def reset_search_path(self, path_type = 'all', no_path = False):
        """Set default search path.

        Parameters
        ----------
            path_type: str
                What type of path to reset. Possibilities are
                'all' (default), 'kdata', 'cia', and 'aerosol'
            no_path: bool (optional)
                If False (default), the search path is reset to '.'.
                If True, empties the search path.
        """
        if no_path:
            local_path=[]
        else:
            local_path=[os.path.abspath('.')]
        if path_type in ('kdata', 'ktable', 'all'):
            self._ktable_search_path = local_path.copy()
        if path_type in ('kdata', 'xtable', 'all'):
            self._xtable_search_path = local_path.copy()
        if path_type in ('cia', 'all'):
            self._cia_search_path = local_path.copy()
        if path_type in ('aerosol', 'all'):
            self._aerosol_search_path = local_path.copy()

    def add_search_path(self, *search_paths, path_type = 'kdata'):
        """Add path(s) to the list of paths that will be searched for
        various files.

        Parameters
        ----------
            search_path : string or list of strings
                Search path(s) to look for opacities.
            path_type: str
                What type of path to change. Possibilities are:
                  * 'kdata' (default): global xsec and corr-k search path
                  * 'ktable' : only corr-k files
                  * 'xtable' : only cross section files
                  * 'cia': search path for CIA files
                  * 'aerosol': search path for Aerosol optical property files

        Examples
        --------

            >>> exo_k.Settings().add_search_path('data/xsec','data/corrk')
            >>> exo_k.Settings().search_path()
            ['/your/path/to/exo_k',
             '/your/path/to/exo_k/data/xsec',
             '/your/path/to/exo_k/data/corrk']
        """
        path_to_change=[]
        if path_type == 'cia':
            path_to_change.append(self._cia_search_path)
        elif path_type == 'aerosol':
            path_to_change.append(self._aerosol_search_path)
        else:
            if path_type in ('kdata', 'ktable'):
                path_to_change.append(self._ktable_search_path)
            if path_type in ('kdata', 'xtable'):
                path_to_change.append(self._xtable_search_path)

        if path_to_change == []:
            raise RuntimeError('I did not understand the path type provided.')
        for path in search_paths:
            if not os.path.isdir(path):
                raise NotADirectoryError("""The search_path you provided
                    does not exist or is not a directory""")
            else:
                for to_change in path_to_change:
                    if os.path.abspath(path) not in to_change:
                        to_change.append(os.path.abspath(path))

    def add_cia_search_path(self, *search_paths):
        """Add path(s) to the list of paths that will be searched for
        cia files.

        Parameters
        ----------
            search_path : string or list of strings
                Search path(s) to look for opacities.
        """
        self.add_search_path(*search_paths, path_type = 'cia')

    def add_aerosol_search_path(self, *search_paths):
        """Add path(s) to the list of paths that will be searched for
        aerosol files.

        Parameters
        ----------
            search_path : string or list of strings
                Search path(s) to look for opacities.
        """
        self.add_search_path(*search_paths, path_type = 'aerosol')

    def set_search_path(self, *search_paths, path_type = 'kdata'):
        """Like :func:~`exo_k.settings.Settings.add_search_path` except for
        the fact that the path is reset first.
        """
        self.reset_search_path(path_type = path_type, no_path = True)
        self.add_search_path(*search_paths, path_type = path_type)

    def set_cia_search_path(self, *search_paths):
        """Like :func:~`exo_k.settings.Settings.add_cia_search_path` except for
        the fact that the path is reset first.
        """
        self.set_search_path(*search_paths, path_type = 'cia')

    def set_aerosol_search_path(self, *search_paths):
        """Like :func:~`exo_k.settings.Settings.add_aerosol_search_path` except for
        the fact that the path is reset first.
        """
        self.set_search_path(*search_paths, path_type = 'aerosol')

    @property
    def search_path(self):
        """Returns the current value of the global search path (_search_path)
        """
        return {'ktable': self._ktable_search_path, 'xtable': self._xtable_search_path,
             'cia': self._cia_search_path, 'aerosol': self._aerosol_search_path}

    def cia_search_path(self):
        """Returns the current value of the cia search path (_cia_search_path)
        """
        raise RuntimeError('Deprecated after v0.0.5. Use xk.Settings().search_path["cia"] instead.')

    def aerosol_search_path(self):
        """Returns the current value of the aerosol search path (_aerosol_search_path)
        """
        raise RuntimeError(
            'Deprecated after v0.0.5. Use xk.Settings().search_path["aerosol"] instead.')

    def set_delimiter(self, newdelimiter):
        """Sets the delimiter string used to separate molecule names in filenames.

        Parameters
        ----------
            newdelimiter: string
                New delimiter to use. Default is '_'.

        Example
        -------
            If I have a file named 'H2O.R10000_xsec.hdf5'
            that I want to load in a `Kdatabase`, the default
            settings will result in an error:

            >>> database=xk.Kdatabase(['H2O'],'R10000')
             No file was found with these filters: 
             ('H2O_', 'R1000') in the following directories:
             ['/home/falco/xsec/xsec_sampled_R10000_0.3-15'] 

            Using

            >>> xk.Settings().set_delimiter('.')
            >>> database=xk.Kdatabase(['H2O'],'R10000')

            finds the file.
        """
        self._delimiter = newdelimiter

    def set_delimiters(self, newdelimiters):
        """Sets the delimiter string used to separate molecule names in filenames.
        If you want to include the '-' character, you must put a '\' before to avoid
        it being interpreted as a special character by the `re` module.

        Parameters
        ----------
            newdelimiters: string
                New delimiters to use. Default is '_.\-'.
        """
        self._delimiters = newdelimiters
        self._search_mol_template = Template('(?=^$mol['+self._delimiters+  \
                ']|['+self._delimiters+']$mol['+self._delimiters+'])')
        self._search_cia_template = Template('(?=^$mol1\-$mol2['+self._delimiters+  \
                ']|^$mol2\-$mol1['+self._delimiters+'])')


    def set_log_interp(self, log_interp):
        """Sets the default interpolation mode for kdata. Default is Log. 

        Parameters
        ----------
            log_interp: boolean
                If True, log interpolation. Linear if False.
        """
        self._log_interp = log_interp

    def set_case_sensitive(self, case_sensitive):
        """Set whether name matching is case sensitive. Default is False.

        Parameters
        ----------
            case_sensitive: boolean
                If True, name matching is case sensitive.
        """
        self._case_sensitive = case_sensitive

    def set_mks(self, set_mks):
        """Forces conversion to mks system.
        
        Parameters
        ----------
            set_mks: boolean
                If True, all datasets are converted to mks upon loading.
        """
        self._convert_to_mks = set_mks

    def list_files(self, *str_filters, molecule = None, 
            only_one = False, search_path = None, path_type = 'kdata'):
        """A routine that provides a list of all filenames containing
        a set of string filters in one of the global _search_path or a local one.

        Whether the search is case sensitive is specified through the
        Settings.set_case_sensitive() method. 

        .. warning::
            The pattern matching with the `str_filters` is done using regular expressions.
            If you want to match special characters (like a dot in a filename), do not
            forget to put a backslash in front of it.

        Parameters
        ----------
            *str_filters: str
                A set of strings that need to be contained in the name of the file
            molecule: str
                The name of a molecule to be looked for in the filename. It must be followed
                by one of the characters in self._delimiters and either at the begining of the name
                or just after one of the characters in self._delimiters.
            only_one: boolean, optional
                If true, only one filename is returned (the first one).
                If false, a list is returned. Default is False.
            search_path: str, optional
                If search_path is provided, it locally overrides
                the global _search_path settings
                and only files in search_path are returned.

        Returns
        -------
            list of strings
                List of filenames corresponding to all the str_filters
        """
        if search_path is not None:
            local_search_path=[search_path]
        else:
            if path_type == 'ktable':
                local_search_path=self._ktable_search_path
            elif path_type == 'xtable':
                local_search_path=self._xtable_search_path
            elif path_type == 'aerosol':
                local_search_path=self._aerosol_search_path

        filenames = [f for path in local_search_path for f in glob(os.path.join(path,'*'))]
        finalnames=filenames[:]
        if molecule is not None:
            if self._case_sensitive:
                template=self._search_mol_template.substitute(mol=molecule)
            else:
                template=self._search_mol_template.substitute(mol=molecule.lower())
        else:
            template='^'
        for filt in str_filters:
            if self._case_sensitive:
                template+='(?=.*'+filt+')'
            else:
                template+='(?=.*'+filt.lower()+')'
        #print(template)
        for filename in filenames:
            if os.path.isdir(filename):
                finalnames.remove(filename)
                continue
            if self._case_sensitive:
                basename=os.path.basename(filename)
            else:
                basename=os.path.basename(filename).lower()
            if re.search(template, basename) is None:
                finalnames.remove(filename)
                continue
        if len(finalnames)>1 and only_one: 
            print("""be careful: {filt}
            String filters not specific enough, several corresponding files have been found.
            We will use the first one:
            {file}
            Other files are:
            {other_files}""".format(filt=str_filters,file=finalnames[0], \
                other_files=finalnames[1:]))
            finalnames=[finalnames[0]]
        if not finalnames:
            raise NoFileFoundError("""No file found to match this regular expression
            (based on your filters): 
            {filt}
            in the following directories:
            {path}
            """.format(filt=template,path=local_search_path))
        # an empty sequence yields False in a conditional statement ! 
        return finalnames

    def list_cia_files(self, *str_filters, molecule_pair = None, 
            only_one = False, search_path = None):
        """A routine that provides a list of all filenames containing
        a set of string filters in the global _search_path or a local one.

        Whether the search is case sensitive is specified through the
        Settings.set_case_sensitive() method. 

        .. warning::
            The pattern matching with the `str_filters` is done using regular expressions.
            If you want to match special characters (like a dot in a filename), do not
            forget to put a backslash in front of it.

        Parameters
        ----------
            *str_filters: str
                A set of strings that need to be contained in the name of the file
            molecule_pair: list of 2 str
                The name of the 2 molecules in the cia pair.
            only_one: boolean, optional
                If true, only one filename is returned (the first one).
                If false, a list is returned. Default is False.
            search_path: str, optional
                If search_path is provided, it locally overrides
                the global _search_path settings
                and only files in search_path are returned.

        Returns
        -------
            list of strings
                List of filenames corresponding to all the str_filters
        """
        if search_path is not None:
            local_search_path=[search_path]
        else:
            local_search_path=self._cia_search_path

        filenames = [f for path in local_search_path for f in glob(os.path.join(path,'*'))]
        finalnames=filenames[:]
        if molecule_pair is not None:
            if self._case_sensitive:
                template=self._search_cia_template.substitute(mol1=molecule_pair[0],
                    mol2=molecule_pair[1])
            else:
                template=self._search_cia_template.substitute(mol1=molecule_pair[0].lower(),
                    mol2=molecule_pair[1].lower())
        else:
            template='^'
        for filt in str_filters:
            if self._case_sensitive:
                template+='(?=.*'+filt+')'
            else:
                template+='(?=.*'+filt.lower()+')'
        #print(template)
        for filename in filenames:
            if os.path.isdir(filename):
                finalnames.remove(filename)
                continue
            if self._case_sensitive:
                basename=os.path.basename(filename)
            else:
                basename=os.path.basename(filename).lower()
            if re.search(template, basename) is None:
                finalnames.remove(filename)
                continue
        if len(finalnames)>1 and only_one: 
            print("""be careful: {filt}
            String filters not specific enough, several corresponding files have been found.
            We will use the first one:
            {file}
            Other files are:
            {other_files}""".format(filt=str_filters,file=finalnames[0], \
                other_files=finalnames[1:]))
            finalnames=[finalnames[0]]
        if not finalnames:
            raise NoFileFoundError("""No file found to match this regular expression
            (based on your filters): 
            {filt}
            in the following directories:
            {path}
            """.format(filt=template,path=local_search_path))
        # an empty sequence yields False in a conditional statement ! 
        return finalnames

class NoFileFoundError(Exception):
    """Error when no file is found
    """
